@@ -19,9 +19,17 @@
  *
  */
 
+/*
+ * ALSA code stolen from http://alsa.opensrc.org/index.php/HowTo_access_a_mixer_control
+ */
+ 
 /* headset-control-obj.c */
 
+#include <stdlib.h>
 #include <string.h>
+#include <alsa/asoundlib.h>
+#include <mce/dbus-names.h>
+#include <mce/mode-names.h>
 #include "headset-control-obj.h"
 #include "hal-marshal.h"
 #include "headset-control-service.h"
@@ -36,8 +44,11 @@
 #define HEADSET_HAL_OBJ_PATH "/org/freedesktop/Hal/devices/computer_logicaldev_input_1"
 #define HAL_DEVICE_INTERFACE "org.freedesktop.Hal.Device"
 
+#define HEADSET_CONTROL_NEXT_MODE_KEY HEADSET_CONTROL_GCONF_FOLDER NEXT_MODE_GCONF_KEY
 #define WANTED_RENDERER "Mafw-Gst-Renderer"
 
+#define IS_IN_CALL g_str_equal (call_state, MCE_CALL_STATE_RINGING) || g_str_equal (call_state, MCE_CALL_STATE_ACTIVE)
+
 #define HEADSET_CONTROL_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
 						HEADSET_TYPE_CONTROL, HeadsetControlPrivate))
 
@@ -47,12 +58,14 @@ struct _HeadsetControlPrivate
 	MafwRenderer *gst_renderer;
 	DBusGConnection *sess_dbus_conn;
 	pb_playback_t *playback;
+	DBusGProxy *mce_signal_proxy;
 	
 	MafwPlayState state;
 	enum pb_state_e new_pb_state;
 
 	gboolean next_mode;
 	gboolean suspended;
+	gboolean in_call;
 
 	guint gconfnotify_id;
 	guint suspend_timer_id;
@@ -60,54 +73,19 @@ struct _HeadsetControlPrivate
 
 enum
 {
-	PROP_0 = 0,
+	PROP_0,
 	PROP_NEXT_MODE,
 	PROP_SUSPENDED,
+	PROP_IN_CALL,
 	PROPCOUNT
 };
 
-static void headset_control_init (HeadsetControl *self); 
-static void headset_control_class_init (HeadsetControlClass *klass); 
-static gpointer headset_control_parent_class = NULL;
-
-static void headset_control_class_intern_init (gpointer klass)
-{
-	headset_control_parent_class = g_type_class_peek_parent (klass);
-	headset_control_class_init (HEADSET_CONTROL_CLASS (klass));
-}
-
-GType headset_control_get_type (void)
-{
-	static volatile gsize g_define_type_id__volatile = 0;
-
-	if (g_once_init_enter (&g_define_type_id__volatile))
-	{ 
-       static GTypeInfo headset_control_info = {
-           sizeof (HeadsetControlClass), /* class size */
-           NULL, /* base_init */
-           NULL, /* base_finalize */
-           (GClassInitFunc) headset_control_class_intern_init, /* class_init */
-           NULL, /* class_finalize */
-           NULL, /* class_data */
-           sizeof (HeadsetControl), /* instance_size */
-           0, /* n_preallocs */
-           (GInstanceInitFunc) headset_control_init, /* instance_init */
-           NULL, /* value_table */ /* Cocaine in the back of the ride */
-       };
-
-		GType g_define_type_id = g_type_register_static (G_TYPE_OBJECT, 
-														 g_intern_static_string ("HeadsetControl"),
-														 &headset_control_info,
-														 (GTypeFlags) 0);
-		g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
-	}
-	
-	return g_define_type_id__volatile;
-}
+#include "headset-control-gtype_shit.inc"
 
 /* -- */
 
 static void headset_control_constructed (GObject *);
+static void headset_control_call_state_changed (DBusGProxy *, const gchar *, const gchar *, HeadsetControl *);
 static void headset_control_button_pressed (DBusGProxy *, const gchar *, const gchar *, HeadsetControl *);
 static void headset_control_renderer_removed (MafwRegistry *, MafwRenderer *, HeadsetControl *);
 static void headset_control_renderer_added_cb (MafwRegistry *, GObject *, HeadsetControl *);
@@ -130,6 +108,10 @@ static void headset_control_get_property (GObject *object, guint property_id, GV
 			g_value_set_boolean (value, self->priv->suspended);
 			break;
 
+		case PROP_IN_CALL:
+			g_value_set_boolean (value, self->priv->in_call);
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
 			break;
@@ -139,6 +121,7 @@ static void headset_control_get_property (GObject *object, guint property_id, GV
 static void headset_control_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec * pspec)
 {
 	HeadsetControl *self = HEADSET_CONTROL (object);
+	HeadsetControlClass *klass = HEADSET_CONTROL_GET_CLASS (self);
  	  	 
 	switch (property_id)
 	{
@@ -151,7 +134,12 @@ static void headset_control_set_property (GObject *object, guint property_id, co
 			self->priv->suspended = g_value_get_boolean (value);
 			g_object_notify (object, "suspended");
 			break;
- 	  	 
+
+		case PROP_IN_CALL:	  	 
+			self->priv->in_call = g_value_get_boolean (value);
+			g_object_notify (object, "in-call");
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
 			break;
@@ -165,11 +153,38 @@ static void headset_control_finalize (GObject *object)
 	HeadsetControl *self = HEADSET_CONTROL (object);
 	HeadsetControlClass *klass = HEADSET_CONTROL_GET_CLASS (self);
 
-	if (klass->gconf_client)
+	DBusGProxy *bus_proxy = dbus_g_proxy_new_for_name (klass->sys_dbus_conn, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
+	if (bus_proxy)
 	{
-		gconf_client_clear_cache (klass->gconf_client);
-		g_object_unref (klass->gconf_client);
-		klass->gconf_client = NULL;
+		org_freedesktop_DBus_release_name (bus_proxy, HEADSET_CONTROL_SERVICE_NAME, NULL, NULL);
+		g_object_unref (G_OBJECT (bus_proxy));
+	}
+
+	headset_control_stop_suspend_off_timer (self);
+
+	self->priv->state = _LastMafwPlayState;
+
+	if (self->priv->mce_signal_proxy)
+	{
+		dbus_g_proxy_disconnect_signal (self->priv->mce_signal_proxy, MCE_CALL_STATE_SIG, G_CALLBACK (headset_control_call_state_changed), self);
+		g_object_unref (G_OBJECT (self->priv->mce_signal_proxy));
+		self->priv->mce_signal_proxy = NULL;
+	}
+
+	if (self->priv->hal_proxy)
+	{
+		dbus_g_proxy_disconnect_signal (self->priv->hal_proxy, HAL_CONDITION, G_CALLBACK (headset_control_button_pressed), self);
+		g_object_unref (G_OBJECT (self->priv->hal_proxy));
+		self->priv->hal_proxy = NULL;
+	}
+
+	g_signal_handlers_disconnect_by_func (klass->mafw_registry, G_CALLBACK (headset_control_renderer_added_cb), self); /* Don't really desire much for renderers to be added... */
+	headset_control_renderer_removed (NULL, self->priv->gst_renderer, self);
+
+	if (self->priv->playback)
+	{
+		pb_playback_destroy (self->priv->playback);
+		self->priv->playback = NULL;
 	}
 
 	if (self->priv->sess_dbus_conn)
@@ -179,6 +194,20 @@ static void headset_control_finalize (GObject *object)
 		self->priv->sess_dbus_conn = NULL;
 	}
 
+	if (self->priv->gconfnotify_id != 0)
+	{
+		gconf_client_notify_remove (klass->gconf_client, self->priv->gconfnotify_id);
+		self->priv->gconfnotify_id = 0;
+		gconf_client_remove_dir (klass->gconf_client, HEADSET_CONTROL_GCONF_FOLDER, NULL);
+	}
+
+	if (klass->gconf_client)
+	{
+		gconf_client_clear_cache (klass->gconf_client);
+		g_object_unref (klass->gconf_client);
+		klass->gconf_client = NULL;
+	}
+
 	mafw_shared_deinit ();
 	klass->mafw_registry = NULL;
 
@@ -188,55 +217,13 @@ static void headset_control_finalize (GObject *object)
 		dbus_g_connection_unref (klass->sys_dbus_conn);
 		klass->sys_dbus_conn = NULL;
 	}
+	
+	self->priv->state = _LastMafwPlayState;
 
 	if (G_OBJECT_CLASS (headset_control_parent_class)->finalize)
 		G_OBJECT_CLASS (headset_control_parent_class)->finalize (object);
 }
 
-static void headset_control_dispose (GObject *object)
-{
-	HeadsetControl *self = HEADSET_CONTROL (object);
-	HeadsetControlClass *klass = HEADSET_CONTROL_GET_CLASS (self);
-	DBusGProxy *bus_proxy;
-
-	headset_control_stop_suspend_off_timer (self);
-
-	if (self->priv->playback)
-	{
-		pb_playback_destroy (self->priv->playback);
-		self->priv->playback = NULL;
-	}
-
-	if (self->priv->gconfnotify_id != 0)
-	{
-		gconf_client_notify_remove (klass->gconf_client, self->priv->gconfnotify_id);
-		self->priv->gconfnotify_id = 0;
-		gconf_client_remove_dir (klass->gconf_client, HEADSET_CONTROL_GCONF_FOLDER, NULL);
-	}
-
-	bus_proxy = dbus_g_proxy_new_for_name (klass->sys_dbus_conn, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
-	if (bus_proxy)
-	{
-		org_freedesktop_DBus_release_name (bus_proxy, HEADSET_CONTROL_SERVICE_NAME, NULL, NULL);
-		g_object_unref (G_OBJECT (bus_proxy));
-	}
-
-	self->priv->state = _LastMafwPlayState;
-
-	g_signal_handlers_disconnect_by_func (klass->mafw_registry, G_CALLBACK (headset_control_renderer_added_cb), self); /* Don't really desire much for renderers to be added... */
-	headset_control_renderer_removed (NULL, self->priv->gst_renderer, self);
-
-	if (self->priv->hal_proxy)
-	{
-		dbus_g_proxy_disconnect_signal (self->priv->hal_proxy, HAL_CONDITION, G_CALLBACK (headset_control_button_pressed), self);
-		g_object_unref (G_OBJECT (self->priv->hal_proxy));
-		self->priv->hal_proxy = NULL;
-	}
-
-	if (G_OBJECT_CLASS (headset_control_parent_class)->dispose)
-		G_OBJECT_CLASS (headset_control_parent_class)->dispose (object);
-}
-
 /* -- */
 
 static void headset_control_class_init (HeadsetControlClass *klass)
@@ -247,7 +234,6 @@ static void headset_control_class_init (HeadsetControlClass *klass)
 
 	object_class->set_property = headset_control_set_property;
 	object_class->get_property = headset_control_get_property;
-	object_class->dispose = headset_control_dispose;
 	object_class->finalize = (GObjectFinalizeFunc) headset_control_finalize;
 	object_class->constructed = headset_control_constructed;
 
@@ -272,6 +258,7 @@ static void headset_control_class_init (HeadsetControlClass *klass)
 
 	g_object_class_install_property (object_class, PROP_NEXT_MODE, g_param_spec_boolean ("next-mode", "Next Mode", "If TRUE, pressing the headset button will skip to the next track, instead.", FALSE, G_PARAM_READWRITE));
 	g_object_class_install_property (object_class, PROP_SUSPENDED, g_param_spec_boolean ("suspended", "60 sec. pause active", "When TRUE, Headset Control will not do anything for 60 seconds when the button is pressed.", FALSE, G_PARAM_READABLE));
+	g_object_class_install_property (object_class, PROP_IN_CALL, g_param_spec_boolean ("in-call", "Self-explanatory, I'd've thought...", "When TRUE, Headset Control will not do anything because the phone is in a call.", FALSE, G_PARAM_READABLE));
 
 	dbus_g_object_type_install_info (HEADSET_TYPE_CONTROL, &dbus_glib_headset_control_object_info);
 	g_type_class_add_private (klass, sizeof (HeadsetControlPrivate));
@@ -292,6 +279,57 @@ static void headset_control_start_suspend_off_timer (HeadsetControl *self)
 	self->priv->suspend_timer_id = g_timeout_add_seconds (60, (GSourceFunc) headset_control_suspend_off, self);
 }
 
+static gboolean headset_control_enable_headset_button (gconstpointer user_data G_GNUC_UNUSED) /* Fuck you, Nokia */
+{
+	gchar *type_contents = NULL;
+	gint type_int = 0;
+	const gchar *card = "hw:0";
+	const gchar *bias_switch = "Jack Bias Switch";
+	gint err = 0;
+	snd_hctl_t *hctl = NULL;
+	snd_ctl_elem_id_t *id;
+	snd_hctl_elem_t *elem;
+	snd_ctl_elem_value_t *control;
+
+	/* Check to see if we have a Nokia headset with a button */
+	if (!g_file_get_contents ("/sys/devices/platform/nokia-av/type", &type_contents, NULL, NULL))
+		goto ret_false;	/* Don't wish to risk anything */
+
+	type_int = atoi (type_contents);
+	g_free (type_contents);
+	
+	if (type_int != 4)
+		goto ret_false;
+
+	err = snd_hctl_open (&hctl, card, 0);
+	if (err < 0)
+		goto ret_false;
+
+	err = snd_hctl_load (hctl);
+	if (err < 0)
+		goto close_hctl;
+
+	snd_ctl_elem_id_alloca (&id);
+	snd_ctl_elem_id_set_interface (id, SND_CTL_ELEM_IFACE_MIXER);
+	snd_ctl_elem_id_set_name (id, bias_switch);
+
+	elem = snd_hctl_find_elem (hctl, id);
+	if (!elem)
+		goto close_hctl;
+
+	snd_ctl_elem_value_alloca (&control);
+	snd_ctl_elem_value_set_id (control, id);
+	snd_ctl_elem_value_set_boolean (control, 0, TRUE);
+	snd_hctl_elem_write (elem, control);
+
+close_hctl:
+	if (hctl)
+		snd_hctl_close (hctl);
+
+ret_false:
+	return FALSE;
+}
+
 /* -- */
 
 static gboolean headset_control_suspend_off (HeadsetControl *self)
@@ -319,6 +357,9 @@ static void headset_control_button_pressed (DBusGProxy *hal_proxy G_GNUC_UNUSED,
 	if (!g_str_equal (condition_name, BUTTON_PRESSED_CONDITION) || !g_str_equal (condition_detail, CONDITION_DETAIL_PHONE))
 		return;
 
+	if (self->priv->in_call)
+		return;
+
 	if (self->priv->suspended)
 	{
 		g_warning ("Sixty Second Pause mode is active!");
@@ -404,6 +445,14 @@ static void headset_control_next_mode_key_changed (GConfClient *gconf_client G_G
 	g_object_set (G_OBJECT (self), "next-mode", gconf_value_get_bool (entry->value), NULL);
 }
 
+static void headset_control_call_state_changed (DBusGProxy *mce_proxy G_GNUC_UNUSED, const gchar *call_state, const gchar *emergency_call_state G_GNUC_UNUSED, HeadsetControl *self)
+{
+	self->priv->in_call = IS_IN_CALL;
+
+	if (!self->priv->in_call)
+		g_timeout_add_seconds (5, (GSourceFunc) headset_control_enable_headset_button, NULL); /* Stupidest fucking hack ever */
+}
+
 /* -- */
 
 static void headset_control_constructed (GObject *object)
@@ -414,6 +463,39 @@ static void headset_control_constructed (GObject *object)
 		G_OBJECT_CLASS (headset_control_parent_class)->constructed (object);
 
 	dbus_g_proxy_connect_signal (self->priv->hal_proxy, HAL_CONDITION, G_CALLBACK (headset_control_button_pressed), self, NULL);
+	(void) headset_control_enable_headset_button (NULL);
+}
+
+static inline void headset_control_mce_phone_watcher_init (HeadsetControl *self)
+{
+	HeadsetControlClass *klass = HEADSET_CONTROL_GET_CLASS (self);
+	DBusGProxy *mce_request_proxy;
+
+	g_return_if_fail (klass->sys_dbus_conn);
+	
+	mce_request_proxy = dbus_g_proxy_new_for_name (klass->sys_dbus_conn, MCE_SERVICE, MCE_REQUEST_PATH, MCE_REQUEST_IF);
+	if (mce_request_proxy)
+	{
+		gchar *call_state = NULL;
+		gchar *emergency_call_state = NULL;
+		
+		if (dbus_g_proxy_call (mce_request_proxy, MCE_CALL_STATE_GET, NULL, G_TYPE_INVALID, G_TYPE_STRING, &call_state, G_TYPE_STRING, &emergency_call_state, G_TYPE_INVALID))
+		{
+			self->priv->in_call = IS_IN_CALL;
+
+			g_free (call_state);
+			g_free (emergency_call_state);
+
+			g_object_unref (G_OBJECT (mce_request_proxy));
+		}
+	}
+
+	self->priv->mce_signal_proxy = dbus_g_proxy_new_for_name (klass->sys_dbus_conn, MCE_SERVICE, MCE_SIGNAL_PATH, MCE_SIGNAL_IF);
+	if (!self->priv->mce_signal_proxy)
+		return;
+
+	dbus_g_proxy_add_signal (self->priv->mce_signal_proxy, MCE_CALL_STATE_SIG, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
+	dbus_g_proxy_connect_signal (self->priv->mce_signal_proxy, MCE_CALL_STATE_SIG, G_CALLBACK (headset_control_call_state_changed), self, NULL);
 }
 
 static inline void headset_control_hal_proxy_init (HeadsetControl *self)
@@ -453,15 +535,12 @@ static inline void headset_control_mafw_renderer_init (HeadsetControl *self)
 
 static inline void headset_control_playback_init (HeadsetControl *self)
 {
-	enum pb_state_e pb_state = PB_STATE_STOP;
+	enum pb_state_e pb_state = self->priv->state == Playing ? PB_STATE_PLAY : PB_STATE_STOP;
 	self->priv->sess_dbus_conn = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);
 	
 	if (!self->priv->sess_dbus_conn)
 		return;
 
-	if (self->priv->state == Playing)
-		pb_state = PB_STATE_PLAY;
-
 	self->priv->playback = pb_playback_new_2 (dbus_g_connection_get_connection (self->priv->sess_dbus_conn), PB_CLASS_MEDIA, PB_FLAG_AUDIO, pb_state, (PBStateRequest) headset_control_libplayback_state_request_handler, NULL);
 }
 
@@ -472,8 +551,7 @@ static inline void headset_control_gconf_init (HeadsetControl *self)
 
 	g_return_if_fail (klass->gconf_client);
 
-	gboolean next_mode = gconf_client_get_bool (klass->gconf_client, HEADSET_CONTROL_GCONF_FOLDER NEXT_MODE_GCONF_KEY, NULL);	
-	self->priv->next_mode = next_mode;
+	gboolean next_mode = gconf_client_get_bool (klass->gconf_client, HEADSET_CONTROL_NEXT_MODE_KEY, NULL);	
 	g_object_set (G_OBJECT (self), "next-mode", next_mode, NULL);
 
 	gconf_client_add_dir (klass->gconf_client, HEADSET_CONTROL_GCONF_FOLDER, GCONF_CLIENT_PRELOAD_NONE, &error);
@@ -484,7 +562,7 @@ static inline void headset_control_gconf_init (HeadsetControl *self)
 		return;
 	}
 
-	self->priv->gconfnotify_id = gconf_client_notify_add (klass->gconf_client, HEADSET_CONTROL_GCONF_FOLDER NEXT_MODE_GCONF_KEY, (GConfClientNotifyFunc) headset_control_next_mode_key_changed, self, NULL, &error);
+	self->priv->gconfnotify_id = gconf_client_notify_add (klass->gconf_client, HEADSET_CONTROL_NEXT_MODE_KEY, (GConfClientNotifyFunc) headset_control_next_mode_key_changed, self, NULL, &error);
 	if (error)
 	{
 		g_warning ("Failed to add next-mode key for watching: %s\n", error->message);
@@ -542,6 +620,7 @@ static void headset_control_init (HeadsetControl *self)
 	headset_control_hal_proxy_init (self);
 	headset_control_mafw_renderer_init (self);
 	headset_control_playback_init (self);
+	headset_control_mce_phone_watcher_init (self);
 	headset_control_dbus_server_init (self);
 }
 
